extern "C" {
#include "rainbow/sha1.h"
}

#include <rainbow/http-client.h>
#include <postgresql/libpq-fe.h>
#include <iostream>
#include <sstream>
#include <roboradio/song-local.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

using namespace std;
using namespace SigC;
using namespace Glib;
using namespace Rainbow;
using namespace Roboradio;

// fixme: PQclear(res) is not called in 4 functions on success

#define IMPROPER_FILENAME "gnomoradio-download"
const string local_file = "/home/garrison/" IMPROPER_FILENAME;

class AvailableCrawl : public SigC::Object
{
public:
	AvailableCrawl (const ustring &url, string &owner, string &_available_id);
	void go ();
	
	Signal1<void,bool> signal_done;
	
private:
	void on_head_done (bool success);
	void on_get_done (bool success);

	void scan_failed ();

	HttpClient *dl;
	ustring url;
	string owner, available_id;
	string file_type;
};

static PGconn *db = 0;

static void connect_to_database ()
{
	if (db && PQstatus(db) == CONNECTION_OK)
		return;
	db = PQconnectdb("host=localhost dbname=gnomoradio");

	if (!db)
		exit(1);
}

static std::string add_slashes (const std::string &input)
{
	char *s = new char[input.size() * 2 + 1];
	PQescapeString(s, input.c_str(), input.size());
	string retval(s);
	delete[] s;
	return retval;
}

static inline char int_to_hex (guint8 x)
{
	// int must be 0-15
	if (x < 10)
		return char(x) + '0';
	else
		return char(x - 10) + 'a';
}

static string audiofile_get_id (string song, string filesize, string checksum, string format, string owner)
{
	bool first_try = true;
	PGresult *res;

 start:
	res = PQexec(db, string("select id from audiofiles where filesize='" + filesize + "' and checksum='" + add_slashes(checksum) + "' and format='" + add_slashes(format) + "' and owner='" + owner + "'").c_str());
	if (PQresultStatus(res) == PGRES_TUPLES_OK
	    && PQntuples(res))
		return PQgetvalue(res, 0, 0);
	PQclear(res);
	
	if (!first_try) {
		cout << "failed to create audiofile\n";
		exit(1);
	}
	first_try = false;

	PQclear(PQexec(db, string("insert into audiofiles (song, filesize, checksum, format, owner) values ('" + song + "', '" + filesize + "', '" + add_slashes(checksum) + "', '" + add_slashes(format) + "', '" + owner + "')").c_str()));
	goto start;
}

static string artist_get_id (string name, string owner)
{
	bool first_try = true;
	PGresult *res;

 start:
	res = PQexec(db, string("select id from artists where name='" + add_slashes(name) + "' and owner='" + owner + "'").c_str());

	if (PQresultStatus(res) == PGRES_TUPLES_OK
	    && PQntuples(res))
		return PQgetvalue(res, 0, 0);
	PQclear(res);

	if (!first_try) {
		cout << "failed to create artist: " << name << "\n";
		exit(1);
	}
	first_try = false;

	PQclear(PQexec(db, string("insert into artists (name, owner) values ('" + add_slashes(name) + "', '" + owner + "')").c_str()));
	goto start;
}

static string album_get_id (string title, string artist, string published, string owner)
{
	bool first_try = true;
	PGresult *res;
	
 start:
	res = PQexec(db, string("select id from albums where title='" + add_slashes(title) + "' and artist='" + artist + "' and owner='" + owner + "'").c_str());
	if (PQresultStatus(res) == PGRES_TUPLES_OK
	    && PQntuples(res))
		return PQgetvalue(res, 0, 0);
	PQclear(res);
	
	if (!first_try) {
		cout << "failed to create album\n";
		exit(1);
	}
	first_try = false;
	
	PQclear(PQexec(db, string("insert into albums (title, artist, published, owner) values ('" + add_slashes(title) + "', '" + artist + "', '" + add_slashes(published) + "', '" + owner + "')").c_str()));
	goto start;
}

static string song_get_id (string title, string artist, string license, unsigned int length, string owner)
{
	bool first_try = true;
	PGresult *res;

	ostringstream length_str;
	length_str << length;
	
 start:
	res = PQexec(db, string("select id from songs where title='" + add_slashes(title) + "' and artist='" + artist + "' and owner='" + owner + "'").c_str());
	if (PQresultStatus(res) == PGRES_TUPLES_OK
	    && PQntuples(res))
		return PQgetvalue(res, 0, 0);
	PQclear(res);
	
	if (!first_try) {
		cout << "failed to create song\n";
		exit(1);
	}
	first_try = false;
	
	PQclear(PQexec(db, string("insert into songs (title, artist, license, length, owner) values ('" + add_slashes(title) + "', '" + artist + "', '" + add_slashes(license) + "', '" + length_str.str() + "', '" + owner + "')").c_str()));
	goto start;
}

AvailableCrawl::AvailableCrawl (const ustring &_url, string &_owner, string &_available_id)
	: url(_url),
	  owner(_owner),
	  available_id(_available_id)
{
}

void AvailableCrawl::go ()
{
	// fixme: check to see if we know about this audio file already?
	
	ustring remote_host, remote_file;
	unsigned short remote_port;
	if (!HttpClient::parse_url(url, remote_host, remote_port, remote_file)) {
		scan_failed();
		signal_done(false);
	}
	dl = new HttpClient(remote_host, remote_port, false);
	dl->signal_request_done.connect(slot(*this, &AvailableCrawl::on_head_done));
	dl->head(remote_file);
}

void AvailableCrawl::on_head_done (bool success)
{
	if (!success) {
		scan_failed();
		signal_done(false);
		delete dl;
		return;
	}

	// check to make sure it is mp3 or ogg
	string h = dl->get_header("content-type");
	if (h == "audio/mp3"
	    || h == "audio/x-mp3"
	    || h == "audio/x-mpeg"
	    || h == "audio/mpeg")
		file_type = "audio/mpeg";
	if (h == "audio/ogg"
	    || h == "audio/x-ogg"
	    || h == "application/ogg"
	    || h == "application/x-ogg")
		file_type = "audio/ogg";
	
	if (file_type == "") {
		// try to determine from extension
		string::size_type p = url.rfind('.');
		if (p != string::npos) {
			string ext = url.substr(p + 1);
			if (ext == "ogg" || ext == "OGG")
				file_type = "audio/ogg";
			else if (ext == "mp3" || ext == "MP3" || ext == "Mp3")
				file_type = "audio/mpeg";
		}
	}
	delete dl;

	ustring remote_host, remote_file;
	unsigned short remote_port;
	HttpClient::parse_url(url, remote_host, remote_port, remote_file);
	dl = new HttpClient(remote_host, remote_port, false);
	dl->signal_request_done.connect(slot(*this, &AvailableCrawl::on_get_done));
	dl->get(remote_file, local_file);
}

void AvailableCrawl::on_get_done (bool success)
{
	if (!success) {
		scan_failed();
		signal_done(false);
		delete dl;
		return;
	}

	SongLocal song(local_file);
	song.obtain_available_info();
	unsigned int length = song.get_length();
	string title = song.get_info("title");
	ustring artist = song.get_info("artist");
	ustring album = song.get_info("album");
	string file_size = dl->get_header("content-length");

	// fallback to filename is title does not exist
	if (title == "" || title == IMPROPER_FILENAME) {
		string::size_type p = url.rfind('/');
		if (p != string::npos) {
			title = url.substr(p + 1);
			string::size_type n = title.rfind('.');
			if (n != string::npos) {
				title = title.substr(0, n);
				for (;;) {
					n = title.find('_');
					if (n == string::npos)
						break;
					title[n] = ' ';
				}
			}
		}
	}

	// calculate checksum
	string xsum;
	FILE *file = fopen(local_file.c_str(), "rb");
	if (file) {
		const int buf_size = 4096;
		unsigned char buf[buf_size];

		SHA1Context cxt;
		SHA1Reset(&cxt);

		size_t r;
		for (;;) {
			r = fread(buf, 1, buf_size, file);
			if (r == 0
			    || SHA1Input(&cxt, buf, r))
				break;
		}
		fclose(file);

		guint8 sum[SHA1HashSize];
		SHA1Result(&cxt, sum);
			
		xsum.resize(SHA1HashSize * 2);
		for (int i = 0; i < SHA1HashSize; ++i) {
			xsum[i*2] = int_to_hex(sum[i] / 16);
			xsum[i*2 + 1] = int_to_hex(sum[i] % 16);
		}
	}

	// make rows or determine ids
	string artist_id = artist_get_id(artist, owner);
	string song_id = song_get_id(title, artist_id, "", length, owner);
	string audiofile_id = audiofile_get_id(song_id, file_size, xsum, file_type, owner);
	if (0 && album.size() != 0) {
		string album_id = album_get_id(album, artist_id, "", owner);
		
		// update albumsongs
		PQclear(PQexec(db, string("insert into albumsongs (song, album) values ('" + song_id + "', '" + album_id + "')").c_str()));
	}

	// make available point to audiofile
	PQclear(PQexec(db, string("update available set audiofile='" + audiofile_id + "' where id='" + available_id + "'").c_str()));
	
	// update last scan
	PQclear(PQexec(db, string("update available set last_scan=current_timestamp where id='" + available_id + "'").c_str()));

	// add to email list
	PQclear(PQexec(db, string("insert into pending_emails (recipient, message) values ('" + owner + "' , 1)").c_str()));

	signal_done(true);
	delete dl;
}

void AvailableCrawl::scan_failed ()
{
	// update last scan
	PQclear(PQexec(db, string("update available set last_scan=current_timestamp where id='" + available_id + "'").c_str()));
}

static void forward (bool success, AvailableCrawl *next)
{
	next->go();
}

static void quit (bool success)
{
	unlink(local_file.c_str());

	exit(success ? 0 : 1);
}

int main ()
{
	struct stat s;
	if (stat(local_file.c_str(), &s) == 0)
		return 1;

	Glib::thread_init();

	cout << "connecting to db..." << endl;
	connect_to_database();
	cout << "connected" << endl;
	
	PGresult *res = PQexec(db, "select id, owner, url, random() as randomorder from available where use='1' and (last_scan='-infinity') order by randomorder");
	if (PQresultStatus(res) == PGRES_TUPLES_OK) {
		cout << PQntuples(res) << " rows" << endl;
		vector<AvailableCrawl*> a;
		a.resize(PQntuples(res));
		for (int i = 0; i < PQntuples(res); ++i) {
			string url = PQgetvalue(res, i, 2);
			string owner = PQgetvalue(res, i, 1);
			string id = PQgetvalue(res, i, 0);
			a[i] = new AvailableCrawl(url, owner, id);
			if (i != 0)
				a[i-1]->signal_done.connect(SigC::bind(slot(&forward), a[i]));
		}
		if (PQntuples(res)) {
			a[PQntuples(res) - 1]->signal_done.connect(slot(&quit));
			a[0]->go();
			Glib::RefPtr<Glib::MainLoop> ml = Glib::MainLoop::create();
			ml->run();
		}
	}

	PQclear(res);

	quit(true);

	return 0;
}
